'''
* ref https://en.wikipedia.org/wiki/Shamir%27s_Secret_Sharing
The essential idea of Adi Shamir's threshold scheme is that 2 points are sufficient to define a line, 3 points are sufficient to define a parabola, 4 points to define a cubic curve and so forth. That is, it takes $k$ points to define a polynomial of degree $k-1$.



'''

import random
from functools import reduce
from operator import add, mul
from klefki.types.algebra.fields import FiniteField
from klefki.types.algebra.meta import field
from klefki.types.algebra.utils import randfield

class SSSS:
    def __init__(self, F: FiniteField):
        self.F = F
        self.node_count=0

    def setup(self, Secret, k, n):
        self.k = k
        self.n = n
        a = [randfield(self.F) for _ in range(k - 1)]
        self.f = lambda x: Secret + reduce(
            add, [a[i] * (x ** i) for i in range(1, k - 1 )])
        return self

    def join(self):
        assert self.node_count < self.n
        self.node_count += 1
        x = randfield(self.F)
        if not hasattr(self, 'f'):
            raise Exception("Needs to encrypt first")
        return (x, self.f(x))


    def decrypt(self, priv):
        x, fx = zip(*priv)
        k = len(fx)
        return reduce(add,
                      [fx[j] * reduce(
                          mul,
                          [x[m] / (x[m]-x[j]) for m in range(k-1) if m != j]) for j in range(k-1)])



from math import gcd
from abc import ABCMeta, abstractproperty
from random import randint
from klefki.types.algebra.utils import randfield
from klefki.types.algebra.meta import field
from klefki.numbers import lcm
import random


class Paillier(metaclass=ABCMeta):

    def __init__(self, P, Q):
        assert gcd(P * Q, (P - 1) * (Q - 1)) == 1

        N = P * Q
        Lam = lcm(P - 1, Q - 1)

        F = field(N)
        DF = field(N**2)

        L  = lambda x: (x - 1) // N
        G = randfield(DF)

        M = ~F(L(pow(G, Lam).value))

        self.privkey = (Lam, M)
        self.pubkey = (N, G)


    @classmethod
    def encrypt(cls, m, pub):
        N, G = pub

        if hasattr(m, "value"):
            m = m.value

        r = G.functor(random.randint(0, N))
        return G**m * r**N


    @classmethod
    def decrypt(cls, c, priv, pub):
        Lam, M = priv
        N, G = pub

        F = M.functor
        L =  lambda x: (x - 1) // N
        return F(L((c ** Lam).value)) * M

    def E(self, m, pub=None):
        return self.encrypt(m, pub or self.pubkey)


    def D(self, c, priv=None, pub=None):
        return self.decrypt(c, priv or self.privkey, pub or self.pubkey)


    def test(self):
        m = random.randint(0, self.N)
        assert m == self.D(self.E(m)).value
        return self
